@@ -20,24 +20,34 @@
     slouken@libsdl.org
 */
 
-/* $Id: SDL_ttf.c,v 1.6 2002/01/18 21:46:04 slouken Exp $ */
+/* $Id: SDL_ttf.c,v 1.15 2003/02/21 17:38:15 slouken Exp $ */
 
 #include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <assert.h>
+
+#ifdef HAVE_ALLOCA_H
+#include <alloca.h>
+#endif
+
+#ifdef HAVE_ALLOCA
+#define ALLOCA(n) ((void*)alloca(n))
+#define FREEA(p)
+#else
+#define ALLOCA(n) malloc(n)
+#define FREEA(p) free(p)
+#endif
 
 #include <freetype/freetype.h>
 #include <freetype/ftoutln.h>
 #include <freetype/ttnameid.h>
+#include <freetype/internal/ftobjs.h>
 
 #include "SDL.h"
+#include "SDL_endian.h"
 #include "SDL_ttf.h"
 
-/* Macro to convert a character to a Unicode value -- assume already Unicode */
-#define UNICODE(c)	c
-
 /* FIXME: Right now we assume the gray-scale renderer Freetype is using
    supports 256 shades of gray, but we should instead key off of num_grays
    in the result FT_Bitmap after the FT_Render_Glyph() call. */
@@ -92,11 +102,65 @@ struct _TTF_Font {
 	c_glyph *current;
 	c_glyph cache[256];
 	c_glyph scratch;
+
+	/* We are responsible for closing the font stream */
+	SDL_RWops *src;
+	int freesrc;
+	FT_Open_Args args;
+
+	/* For non-scalable formats, we must remember which font index size */
+	int font_size_family;
 };
 
 /* The FreeType font engine/library */
 static FT_Library library;
 static int TTF_initialized = 0;
+static int TTF_byteswapped = 0;
+
+/* UNICODE string utilities */
+static __inline__ int UNICODE_strlen(const Uint16 *text)
+{
+	int size = 0;
+	while ( *text++ ) {
+		++size;
+	}
+	return size;
+}
+static __inline__ void UNICODE_strcpy(Uint16 *dst, const Uint16 *src, int swap)
+{
+	if ( swap ) {
+		while ( *src ) {
+			*dst = SDL_Swap16(*src);
+			++src;
+			++dst;
+		}
+		*dst = '\0';
+	} else {
+		while ( *src ) {
+			*dst = *src;
+			++src;
+			++dst;
+		}
+		*dst = '\0';
+	}
+}
+
+/* rcg06192001 get linked library's version. */
+const SDL_version *TTF_Linked_Version(void)
+{
+	static SDL_version linked_version;
+	TTF_VERSION(&linked_version);
+	return(&linked_version);
+}
+
+/* This function tells the library whether UNICODE text is generally
+   byteswapped.  A UNICODE BOM character at the beginning of a string
+   will override this setting for that string.
+ */
+void TTF_ByteSwappedUNICODE(int swapped)
+{
+	TTF_byteswapped = swapped;
+}
 
 static void TTF_SetFTError(const char *msg, FT_Error error)
 {
@@ -133,84 +197,141 @@ static void TTF_SetFTError(const char *msg, FT_Error error)
 
 int TTF_Init( void )
 {
-	int status;
-	FT_Error error;
+	int status = 0;
 
-	status = 0;
-	error = FT_Init_FreeType( &library );
-	if ( error ) {
-		TTF_SetFTError( "Couldn't init FreeType engine", error );
-		status = -1;
-	} else {
-		TTF_initialized = 1;
+	if ( ! TTF_initialized ) {
+		FT_Error error = FT_Init_FreeType( &library );
+		if ( error ) {
+			TTF_SetFTError("Couldn't init FreeType engine", error);
+			status = -1;
+		}
+	}
+	if ( status == 0 ) {
+		++TTF_initialized;
 	}
 	return status;
 }
 
-TTF_Font* TTF_OpenFontIndex( const char *file, int ptsize, long index )
+static unsigned long RWread(
+	FT_Stream stream,
+	unsigned long offset,
+	unsigned char* buffer,
+	unsigned long count
+)
+{
+	SDL_RWops *src;
+
+	src = (SDL_RWops *)stream->descriptor.pointer;
+	SDL_RWseek( src, (int)offset, SEEK_SET );
+	return SDL_RWread( src, buffer, 1, (int)count );
+}
+
+TTF_Font* TTF_OpenFontIndexRW( SDL_RWops *src, int freesrc, int ptsize, long index )
 {
 	TTF_Font* font;
 	FT_Error error;
 	FT_Face face;
 	FT_Fixed scale;
+	FT_Stream stream;
+	int position;
+
+	if ( ! TTF_initialized ) {
+		TTF_SetError( "Library not initialized" );
+		return NULL;
+	}
+
+	/* Check to make sure we can seek in this stream */
+	position = SDL_RWtell(src);
+	if ( position < 0 ) {
+		TTF_SetError( "Can't seek in stream" );
+		return NULL;
+	}
 
 	font = (TTF_Font*) malloc(sizeof *font);
 	if ( font == NULL ) {
 		TTF_SetError( "Out of memory" );
 		return NULL;
 	}
-	memset( font, 0, sizeof( *font ) );
+	memset(font, 0, sizeof(*font));
 
-	/* Open the font and create ancillary data */
-	error = FT_New_Face( library, file, 0, &font->face );
-	if( error ) {
-		TTF_SetFTError( "Couldn't load font file", error);
-		free( font );
+	font->src = src;
+	font->freesrc = freesrc;
+
+	stream = (FT_Stream)malloc(sizeof(*stream));
+	if ( stream == NULL ) {
+		TTF_SetError( "Out of memory" );
+		TTF_CloseFont( font );
 		return NULL;
 	}
-	if ( index != 0 ) {
-		if ( font->face->num_faces > index ) {
-		  	FT_Done_Face( font->face );
-			error = FT_New_Face( library, file, index, &font->face );
-			if( error ) {
-				TTF_SetFTError( "Couldn't get font face", error);
-				free( font );
-				return NULL;
-			}
-		} else {
-			TTF_SetFTError( "No such font face", error);
-			free( font );
-			return NULL;
-		}
+	memset(stream, 0, sizeof(*stream));
+
+	stream->memory = library->memory;
+	stream->read = RWread;
+	stream->descriptor.pointer = src;
+	stream->pos = (unsigned long)position;
+	SDL_RWseek(src, 0, SEEK_END);
+	stream->size = (unsigned long)(SDL_RWtell(src) - position);
+	SDL_RWseek(src, position, SEEK_SET);
+
+	font->args.flags = ft_open_stream;
+	font->args.stream = stream;
+
+	error = FT_Open_Face( library, &font->args, index, &font->face );
+	if( error ) {
+		TTF_SetFTError( "Couldn't load font file", error );
+		TTF_CloseFont( font );
+		return NULL;
 	}
 	face = font->face;
 
 	/* Make sure that our font face is scalable (global metrics) */
-	if ( ! FT_IS_SCALABLE(face) ) {
-		TTF_SetError("Font face is not scalable");
-		TTF_CloseFont( font );
-		return NULL;
+	if ( FT_IS_SCALABLE(face) ) {
+
+	  	/* Set the character size and use default DPI (72) */
+	  	error = FT_Set_Char_Size( font->face, 0, ptsize * 64, 0, 0 );
+			if( error ) {
+	    	TTF_SetFTError( "Couldn't set font size", error );
+	    	TTF_CloseFont( font );
+	    	return NULL;
+	  }
+
+	  /* Get the scalable font metrics for this font */
+	  scale = face->size->metrics.y_scale;
+	  font->ascent  = FT_CEIL(FT_MulFix(face->bbox.yMax, scale));
+	  font->descent = FT_CEIL(FT_MulFix(face->bbox.yMin, scale));
+	  font->height  = font->ascent - font->descent + /* baseline */ 1;
+	  font->lineskip = FT_CEIL(FT_MulFix(face->height, scale));
+	  font->underline_offset = FT_FLOOR(FT_MulFix(face->underline_position, scale));
+	  font->underline_height = FT_FLOOR(FT_MulFix(face->underline_thickness, scale));
+
+	} else {
+		/* Non-scalable font case.  ptsize determines which family
+		 * or series of fonts to grab from the non-scalable format.
+		 * It is not the point size of the font.
+		 * */
+		if ( ptsize >= font->face->num_fixed_sizes )
+			ptsize = font->face->num_fixed_sizes - 1;
+		font->font_size_family = ptsize;
+		error = FT_Set_Pixel_Sizes( face,
+				face->available_sizes[ptsize].height,
+				face->available_sizes[ptsize].width );
+	  	/* With non-scalale fonts, Freetype2 likes to fill many of the
+		 * font metrics with the value of 0.  The size of the
+		 * non-scalable fonts must be determined differently
+		 * or sometimes cannot be determined.
+		 * */
+	  	font->ascent = face->available_sizes[ptsize].height;
+	  	font->descent = 0;
+	  	font->height = face->available_sizes[ptsize].height;
+	  	font->lineskip = FT_CEIL(font->ascent);
+	  	font->underline_offset = FT_FLOOR(face->underline_position);
+	  	font->underline_height = FT_FLOOR(face->underline_thickness);
 	}
 
-	/* Set the character size and use default DPI (72) */
-	error = FT_Set_Char_Size( font->face, 0, ptsize * 64, 0, 0 );
-	if( error ) {
-		TTF_SetFTError( "Couldn't set font size", error );
-		TTF_CloseFont( font );
-		return NULL;
-	}
-
-	/* Get the scalable font metrics for this font */
-	scale = face->size->metrics.y_scale;
-	font->ascent  = FT_CEIL(FT_MulFix(face->bbox.yMax, scale));
-	font->descent = FT_CEIL(FT_MulFix(face->bbox.yMin, scale));
-	font->height  = font->ascent - font->descent + /* baseline */ 1;
-	font->lineskip = FT_CEIL(FT_MulFix(face->height, scale));
-	font->underline_offset = FT_FLOOR(FT_MulFix(face->underline_position, scale));
-	font->underline_height = FT_FLOOR(FT_MulFix(face->underline_thickness, scale));
 	if ( font->underline_height < 1 ) {
 		font->underline_height = 1;
 	}
+
 #ifdef DEBUG_FONTS
 	printf("Font metrics:\n");
 	printf("\tascent = %d, descent = %d\n",
@@ -231,6 +352,16 @@ TTF_Font* TTF_OpenFontIndex( const char *file, int ptsize, long index )
 	return font;
 }
 
+TTF_Font* TTF_OpenFontRW( SDL_RWops *src, int freesrc, int ptsize )
+{
+	return TTF_OpenFontIndexRW(src, freesrc, ptsize, 0);
+}
+
+TTF_Font* TTF_OpenFontIndex( const char *file, int ptsize, long index )
+{
+	return TTF_OpenFontIndexRW(SDL_RWFromFile(file, "rb"), 1, ptsize, index);
+}
+
 TTF_Font* TTF_OpenFont( const char *file, int ptsize )
 {
 	return TTF_OpenFontIndex(file, ptsize, 0);
@@ -276,8 +407,9 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 	FT_Glyph_Metrics* metrics;
 	FT_Outline* outline;
 
-	assert( font );
-	assert( font->face );
+	if ( !font || !font->face ) {
+		return FT_Err_Invalid_Handle;
+	}
 
 	face = font->face;
 
@@ -297,14 +429,29 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 
 	/* Get the glyph metrics if desired */
 	if ( (want & CACHED_METRICS) && !(cached->stored & CACHED_METRICS) ) {
-		/* Get the bounding box */
-		cached->minx = FT_FLOOR(metrics->horiBearingX);
-		cached->maxx = cached->minx + FT_CEIL(metrics->width);
-		cached->maxy = FT_FLOOR(metrics->horiBearingY);
-		cached->miny = cached->maxy - FT_CEIL(metrics->height);
-		cached->yoffset = font->ascent - cached->maxy;
-		cached->advance = FT_CEIL(metrics->horiAdvance);
-
+		if ( FT_IS_SCALABLE( face ) ) {
+			/* Get the bounding box */
+			cached->minx = FT_FLOOR(metrics->horiBearingX);
+			cached->maxx = cached->minx + FT_CEIL(metrics->width);
+			cached->maxy = FT_FLOOR(metrics->horiBearingY);
+			cached->miny = cached->maxy - FT_CEIL(metrics->height);
+			cached->yoffset = font->ascent - cached->maxy;
+			cached->advance = FT_CEIL(metrics->horiAdvance);
+		} else {
+			/* Get the bounding box for non-scalable format.
+			 * Again, freetype2 fills in many of the font metrics
+			 * with the value of 0, so some of the values we
+			 * need must be calculated differently with certain
+			 * assumptions about non-scalable formats.
+			 * */
+			cached->minx = FT_FLOOR(metrics->horiBearingX);
+			cached->maxx = cached->minx + FT_CEIL(metrics->horiAdvance);
+			cached->maxy = FT_FLOOR(metrics->horiBearingY);
+			cached->miny = cached->maxy - FT_CEIL(face->available_sizes[font->font_size_family].height);
+			cached->yoffset = 0;
+			cached->advance = FT_CEIL(metrics->horiAdvance);
+		}
+		
 		/* Adjust for bold and italic text */
 		if( font->style & TTF_STYLE_BOLD ) {
 			cached->maxx += font->glyph_overhang;
@@ -352,7 +499,14 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 			dst = &cached->pixmap;
 		}
 		memcpy( dst, src, sizeof( *dst ) );
-		if ( mono ) {
+
+		/* FT_Render_Glyph() and .fon fonts always generate a
+		 * two-color (black and white) glyphslot surface, even
+		 * when rendered in ft_render_mode_normal.  This is probably
+		 * a freetype2 bug because it is inconsistent with the
+		 * freetype2 documentation under FT_Render_Mode section.
+		 * */
+		if ( mono || !FT_IS_SCALABLE(face) ) {
 			dst->pitch *= 8;
 		}
 
@@ -400,6 +554,31 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 						ch <<= 1;
 						*dstp++ = (ch&0x80) >> 7;
 					}
+				} else if ( !FT_IS_SCALABLE(face) ) {
+					/* This special case wouldn't
+					 * be here if the FT_Render_Glyph()
+					 * function wasn't buggy when it tried
+					 * to render a .fon font with 256
+					 * shades of gray.  Instead, it
+					 * returns a black and white surface
+					 * and we have to translate it back
+					 * to a 256 gray shaded surface. 
+					 * */
+					unsigned char *srcp = src->buffer + soffset;
+					unsigned char *dstp = dst->buffer + doffset;
+					unsigned char ch;
+					int j, k;
+					for ( j = 0; j < src->width; j += 8) {
+						ch = *srcp++;
+						for (k = 0; k < 8; ++k) {
+							if ((ch&0x80) >> 7) {
+								*dstp++ = NUM_GRAYS - 1;
+							} else {
+								*dstp++ = 0x00;
+							}
+							ch <<= 1;
+						}
+					}
 				} else {
 					memcpy(dst->buffer+doffset,
 					       src->buffer+soffset, src->pitch);
@@ -465,11 +644,19 @@ static FT_Error Find_Glyph( TTF_Font* font, Uint16 ch, int want )
 void TTF_CloseFont( TTF_Font* font )
 {
 	Flush_Cache( font );
-	FT_Done_Face( font->face );
+	if ( font->face ) {
+		FT_Done_Face( font->face );
+	}
+	if ( font->args.stream ) {
+		free( font->args.stream );
+	}
+	if ( font->freesrc ) {
+		SDL_RWclose( font->src );
+	}
 	free( font );
 }
 
-static Uint16 *ASCII_to_UNICODE(Uint16 *unicode, const char *text, int len)
+static Uint16 *LATIN1_to_UNICODE(Uint16 *unicode, const char *text, int len)
 {
 	int i;
 
@@ -587,18 +774,19 @@ int TTF_SizeText(TTF_Font *font, const char *text, int *w, int *h)
 
 	/* Copy the Latin-1 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return -1;
 	}
-	ASCII_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	LATIN1_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	status = TTF_SizeUNICODE(font, unicode_text, w, h);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return status;
 }
 
@@ -610,18 +798,19 @@ int TTF_SizeUTF8(TTF_Font *font, const char *text, int *w, int *h)
 
 	/* Copy the UTF-8 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return -1;
 	}
-	UTF8_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	UTF8_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	status = TTF_SizeUNICODE(font, unicode_text, w, h);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return status;
 }
 
@@ -629,6 +818,7 @@ int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)
 {
 	int status;
 	const Uint16 *ch;
+	int swapped;
 	int x, z;
 	int minx, maxx;
 	int miny, maxy;
@@ -637,21 +827,60 @@ int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)
 
 	/* Initialize everything to 0 */
 	if ( ! TTF_initialized ) {
+		TTF_SetError( "Library not initialized" );
 		return -1;
 	}
 	status = 0;
 	minx = maxx = 0;
 	miny = maxy = 0;
+	swapped = TTF_byteswapped;
 
 	/* Load each character and sum it's bounding box */
 	x= 0;
 	for ( ch=text; *ch; ++ch ) {
-		error = Find_Glyph(font, *ch, CACHED_METRICS);
+		Uint16 c = *ch;
+		if ( c == UNICODE_BOM_NATIVE ) {
+			swapped = 0;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( c == UNICODE_BOM_SWAPPED ) {
+			swapped = 1;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( swapped ) {
+			c = SDL_Swap16(c);
+		}
+
+		error = Find_Glyph(font, c, CACHED_METRICS);
 		if ( error ) {
 			return -1;
 		}
 		glyph = font->current;
 
+		if ( (ch == text) && (glyph->minx < 0) ) {
+		/* Fixes the texture wrapping bug when the first letter
+		 * has a negative minx value or horibearing value.  The entire
+		 * bounding box must be adjusted to be bigger so the entire
+		 * letter can fit without any texture corruption or wrapping.
+		 *
+		 * Effects: First enlarges bounding box.
+		 * Second, xstart has to start ahead of its normal spot in the
+		 * negative direction of the negative minx value.
+		 * (pushes everything to the right).
+		 *
+		 * This will make the memory copy of the glyph bitmap data
+		 * work out correctly.
+		 * */
+			z -= glyph->minx;
+			
+		}
+		
 		z = x + glyph->minx;
 		if ( minx > z ) {
 			minx = z;
@@ -702,18 +931,19 @@ SDL_Surface *TTF_RenderText_Solid(TTF_Font *font,
 
 	/* Copy the Latin-1 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return(NULL);
 	}
-	ASCII_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	LATIN1_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	textbuf = TTF_RenderUNICODE_Solid(font, unicode_text, fg);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return(textbuf);
 }
 
@@ -728,18 +958,19 @@ SDL_Surface *TTF_RenderUTF8_Solid(TTF_Font *font,
 
 	/* Copy the UTF-8 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return(NULL);
 	}
+	*unicode_text = UNICODE_BOM_NATIVE;
 	UTF8_to_UNICODE(unicode_text, text, unicode_len);
 
 	/* Render the new text */
 	textbuf = TTF_RenderUNICODE_Solid(font, unicode_text, fg);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return(textbuf);
 }
 
@@ -754,9 +985,10 @@ SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
 	const Uint16* ch;
 	Uint8* src;
 	Uint8* dst;
-	int row;
-	int col;
+	int swapped;
+	int row, col;
 	c_glyph *glyph;
+	FT_Bitmap *current;
 	FT_Error error;
 
 	/* Get the dimensions of the text surface */
@@ -784,19 +1016,44 @@ SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
 
 	/* Load and render each character */
 	xstart = 0;
-
+	swapped = TTF_byteswapped;
 	for( ch=text; *ch; ++ch ) {
-		FT_Bitmap* current = NULL;
+		Uint16 c = *ch;
+		if ( c == UNICODE_BOM_NATIVE ) {
+			swapped = 0;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( c == UNICODE_BOM_SWAPPED ) {
+			swapped = 1;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( swapped ) {
+			c = SDL_Swap16(c);
+		}
 
-		error = Find_Glyph(font, *ch, CACHED_METRICS|CACHED_BITMAP);
+		error = Find_Glyph(font, c, CACHED_METRICS|CACHED_BITMAP);
 		if( error ) {
 			SDL_FreeSurface( textbuf );
 			return NULL;
 		}
 		glyph = font->current;
-
 		current = &glyph->bitmap;
+		/* Compensate for wrap around bug with negative minx's */
+		if ( (ch == text) && (glyph->minx < 0) ) {
+			xstart -= glyph->minx;
+		}
+		
 		for( row = 0; row < current->rows; ++row ) {
+			/* Make sure we don't go over the limit */
+			if ( row+glyph->yoffset >= textbuf->h ) {
+				continue;
+			}
 			dst = (Uint8*) textbuf->pixels +
 				(row+glyph->yoffset) * textbuf->pitch +
 				xstart + glyph->minx;
@@ -847,8 +1104,8 @@ SDL_Surface *TTF_RenderGlyph_Solid(TTF_Font *font, Uint16 ch, SDL_Color fg)
 
 	/* Create the target surface */
 	textbuf = SDL_CreateRGBSurface( SDL_SWSURFACE,
-					glyph->pixmap.width,
-					glyph->pixmap.rows,
+					glyph->bitmap.pitch,
+					glyph->bitmap.rows,
 					8, 0, 0, 0, 0 );
 	if ( ! textbuf ) {
 		return(NULL);
@@ -865,11 +1122,11 @@ SDL_Surface *TTF_RenderGlyph_Solid(TTF_Font *font, Uint16 ch, SDL_Color fg)
 	SDL_SetColorKey(textbuf, SDL_SRCCOLORKEY, 0);
 
 	/* Copy the character from the pixmap */
-	src = glyph->pixmap.buffer;
+	src = glyph->bitmap.buffer;
 	dst = (Uint8*) textbuf->pixels;
 	for ( row = 0; row < textbuf->h; ++row ) {
-		memcpy( dst, src, glyph->pixmap.pitch );
-		src += glyph->pixmap.pitch;
+		memcpy( dst, src, glyph->bitmap.pitch );
+		src += glyph->bitmap.pitch;
 		dst += textbuf->pitch;
 	}
 
@@ -901,18 +1158,19 @@ SDL_Surface *TTF_RenderText_Shaded(TTF_Font *font,
 
 	/* Copy the Latin-1 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return(NULL);
 	}
-	ASCII_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	LATIN1_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	textbuf = TTF_RenderUNICODE_Shaded(font, unicode_text, fg, bg);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return(textbuf);
 }
 
@@ -927,18 +1185,19 @@ SDL_Surface *TTF_RenderUTF8_Shaded(TTF_Font *font,
 
 	/* Copy the UTF-8 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return(NULL);
 	}
-	UTF8_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	UTF8_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	textbuf = TTF_RenderUNICODE_Shaded(font, unicode_text, fg, bg);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return(textbuf);
 }
 
@@ -959,7 +1218,9 @@ SDL_Surface* TTF_RenderUNICODE_Shaded( TTF_Font* font,
 	const Uint16* ch;
 	Uint8* src;
 	Uint8* dst;
+	int swapped;
 	int row, col;
+	FT_Bitmap* current;
 	c_glyph *glyph;
 	FT_Error error;
 
@@ -990,18 +1251,44 @@ SDL_Surface* TTF_RenderUNICODE_Shaded( TTF_Font* font,
 
 	/* Load and render each character */
 	xstart = 0;
+	swapped = TTF_byteswapped;
 	for( ch = text; *ch; ++ch ) {
-		FT_Bitmap* current;
+		Uint16 c = *ch;
+		if ( c == UNICODE_BOM_NATIVE ) {
+			swapped = 0;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( c == UNICODE_BOM_SWAPPED ) {
+			swapped = 1;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( swapped ) {
+			c = SDL_Swap16(c);
+		}
 
-		error = Find_Glyph(font, *ch, CACHED_METRICS|CACHED_PIXMAP);
+		error = Find_Glyph(font, c, CACHED_METRICS|CACHED_PIXMAP);
 		if( error ) {
 			SDL_FreeSurface( textbuf );
 			return NULL;
 		}
 		glyph = font->current;
-
+		/* Compensate for the wrap around with negative minx's */
+		if ( (ch == text) && (glyph->minx < 0) ) {
+			xstart -= glyph->minx;
+		}
+		
 		current = &glyph->pixmap;
 		for( row = 0; row < current->rows; ++row ) {
+			/* Make sure we don't go over the limit */
+			if ( row+glyph->yoffset >= textbuf->h ) {
+				continue;
+			}
 			dst = (Uint8*) textbuf->pixels +
 				(row+glyph->yoffset) * textbuf->pitch +
 				xstart + glyph->minx;
@@ -1111,18 +1398,19 @@ SDL_Surface *TTF_RenderText_Blended(TTF_Font *font,
 
 	/* Copy the Latin-1 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return(NULL);
 	}
-	ASCII_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	LATIN1_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	textbuf = TTF_RenderUNICODE_Blended(font, unicode_text, fg);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return(textbuf);
 }
 
@@ -1137,18 +1425,19 @@ SDL_Surface *TTF_RenderUTF8_Blended(TTF_Font *font,
 
 	/* Copy the UTF-8 text to a UNICODE text buffer */
 	unicode_len = strlen(text);
-	unicode_text = (Uint16 *)malloc((unicode_len+1)*(sizeof *unicode_text));
+	unicode_text = (Uint16 *)ALLOCA((1+unicode_len+1)*(sizeof *unicode_text));
 	if ( unicode_text == NULL ) {
 		TTF_SetError("Out of memory");
 		return(NULL);
 	}
-	UTF8_to_UNICODE(unicode_text, text, unicode_len);
+	*unicode_text = UNICODE_BOM_NATIVE;
+	UTF8_to_UNICODE(unicode_text+1, text, unicode_len);
 
 	/* Render the new text */
 	textbuf = TTF_RenderUNICODE_Blended(font, unicode_text, fg);
 
 	/* Free the text buffer and return */
-	free(unicode_text);
+	FREEA(unicode_text);
 	return(textbuf);
 }
 
@@ -1163,6 +1452,7 @@ SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
 	const Uint16 *ch;
 	Uint8 *src;
 	Uint32 *dst;
+	int swapped;
 	int row, col;
 	c_glyph *glyph;
 	FT_Error error;
@@ -1182,22 +1472,52 @@ SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
 
 	/* Load and render each character */
 	xstart = 0;
+	swapped = TTF_byteswapped;
 	pixel = (fg.r<<16)|(fg.g<<8)|fg.b;
 	for ( ch=text; *ch; ++ch ) {
-		error = Find_Glyph(font, *ch, CACHED_METRICS|CACHED_PIXMAP);
+		Uint16 c = *ch;
+		if ( c == UNICODE_BOM_NATIVE ) {
+			swapped = 0;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( c == UNICODE_BOM_SWAPPED ) {
+			swapped = 1;
+			if ( text == ch ) {
+				++text;
+			}
+			continue;
+		}
+		if ( swapped ) {
+			c = SDL_Swap16(c);
+		}
+		error = Find_Glyph(font, c, CACHED_METRICS|CACHED_PIXMAP);
 		if( error ) {
 			SDL_FreeSurface( textbuf );
 			return NULL;
 		}
 		glyph = font->current;
-
 		width = glyph->pixmap.width;
-		src = (Uint8 *)glyph->pixmap.buffer;
+		/* Compensate for the wrap around bug with negative minx's */
+		if ( (ch == text) && (glyph->minx < 0) ) {
+			xstart -= glyph->minx;
+		}
+
 		for ( row = 0; row < glyph->pixmap.rows; ++row ) {
+			/* Make sure we don't go over the limit */
+			if ( row+glyph->yoffset >= textbuf->h ) {
+				continue;
+			}
 			dst = (Uint32*) textbuf->pixels +
 				(row+glyph->yoffset) * textbuf->pitch/4 +
 				xstart + glyph->minx;
-			for ( col=width; col>0; --col ) {
+			/* Added code to adjust src pointer for pixmaps to
+			 * account for pitch.
+			 * */
+			src = (Uint8*) (glyph->pixmap.buffer + glyph->pixmap.pitch * row);
+			for ( col = width; col>0; --col) {
 				alpha = *src++;
 				*dst++ |= pixel | (alpha << 24);
 			}
@@ -1255,7 +1575,8 @@ SDL_Surface *TTF_RenderGlyph_Blended(TTF_Font *font, Uint16 ch, SDL_Color fg)
 	/* Copy the character from the pixmap */
 	pixel = (fg.r<<16)|(fg.g<<8)|fg.b;
 	for ( row=0; row<textbuf->h; ++row ) {
-		src = glyph->pixmap.buffer + row * glyph->pixmap.width;
+		/* Changed src to take pitch into account, not just width */
+		src = glyph->pixmap.buffer + row * glyph->pixmap.pitch;
 		dst = (Uint32 *)textbuf->pixels + row * textbuf->pitch/4;
 		for ( col=0; col<glyph->pixmap.width; ++col ) {
 			alpha = *src++;
@@ -1295,7 +1616,13 @@ int TTF_GetFontStyle( TTF_Font* font )
 void TTF_Quit( void )
 {
 	if ( TTF_initialized ) {
-		FT_Done_FreeType( library );
+		if ( --TTF_initialized == 0 ) {
+			FT_Done_FreeType( library );
+		}
 	}
-	TTF_initialized = 0;
+}
+
+int TTF_WasInit( void )
+{
+	return TTF_initialized;
 }
